// adapt_dealii2ublas.h -*-c++-*-
#ifndef _adapt_dealii2ublas_HDR_
#define _adapt_dealii2ublas_HDR_

#define BOOST_UBLAS_SHALLOW_ARRAY_ADAPTOR
#include<boost/numeric/ublas/storage.hpp> // for shallow_array_adaptor
#include<boost/numeric/ublas/vector.hpp>

#include<lac/vector.h>

namespace pu{
  template<typename V> class AdaptToUblas;
  template<typename SC> class AdaptToUblas<Vector<SC> >{
  private:
    typedef boost::numeric::ublas::shallow_array_adaptor<SC> adaptor_t;
  public:
    typedef boost::numeric::ublas::vector<SC,adaptor_t> vector_type;
  private:
    adaptor_t a;
    vector_type v;
  public:
    AdaptToUblas(Vector<SC> &VAdapted)
      :a(VAdapted.size(),&*(VAdapted.begin())),
       v(VAdapted.size(),a)
    {}
    vector_type & get_vector() { return v;
    }
    vector_type const & get_vector() const { return v;
    }
  };
  template<typename SC> class AdaptToUblas<Vector<SC> const>{
  private:

    //
    // SC is const
    // to not to do const_cast<SC *> at construction of shallow_array_adaptor
    // as it can be constructed only with "pointer", and not "const pointer".
    //
    typedef boost::numeric::ublas::shallow_array_adaptor<const SC> adaptor_t;

  public:
    typedef const boost::numeric::ublas::vector<SC /*const*/,
						adaptor_t const> vector_type;
  private:
    adaptor_t const a;
    vector_type const v;
  public:
    AdaptToUblas(Vector<SC> const &VAdapted)
      :a(VAdapted.size(),&*(VAdapted.begin())),
       v(VAdapted.size(),a)
    {}
    vector_type const & get_vector() const { return v;
    }
  };
}//of pu
#endif
