// addins_for_num_bindings.hpp -*-c++-*-
#ifndef addins_for_num_bindings_HDR_
#define addins_for_num_bindings_HDR_

#include<boost/numeric/bindings/traits/vector_traits.hpp>
#include<boost/numeric/bindings/traits/sparse_traits.hpp>

namespace boost {
  namespace numeric{
    namespace bindings {
      namespace traits  {

	namespace detail {
	  template<typename T>
	  struct non_const_type_resolve{
	    typedef T type;
	  };
	  template<typename T>
	  struct non_const_type_resolve<const T>{
	    typedef T type;
	  };
	}

	template<typename SC,typename INDEX_T=std::size_t>
	struct typical_vector_bridge {
	  //for const vectors, define it with const SC, same effect
	  typedef SC value_type;
	  typedef SC *pointer;
	  typedef const SC *const_pointer;

	  pointer const storage_beg;
	  INDEX_T const stride;
	  INDEX_T const size;
	  
	public:
	  typical_vector_bridge(pointer const beg,INDEX_T const stride_in,
				INDEX_T const size_in)
	    :storage_beg(beg),stride(stride_in),size(size_in)
	  {}
	  template<typename VECTOR_T> explicit
	  typical_vector_bridge(VECTOR_T &v)
	    :storage_beg(boost::numeric::bindings::traits::vector_storage(v)),
	     stride(boost::numeric::bindings::traits::vector_stride(v)),
	     size(boost::numeric::bindings::traits::vector_size(v))
	  {}
	};

	template<typename SC,typename INDEX_T=std::size_t>
	struct typical_const_vector_bridge {
	  //for const vectors, define it with const SC, same effect
	  typedef SC value_type;
	  typedef SC const *pointer;
	  typedef const SC *const_pointer;

	  pointer const storage_beg;
	  INDEX_T const stride;
	  INDEX_T const size;
	  
	public:
	  typical_const_vector_bridge
	  (pointer const beg,INDEX_T const stride_in,INDEX_T const size_in)
	    :storage_beg(beg),stride(stride_in),size(size_in)
	  {}
	  template<typename VECTOR_T> explicit
	  typical_const_vector_bridge(VECTOR_T &v)
	    :storage_beg(boost::numeric::bindings::traits::vector_storage(v)),
	     stride(boost::numeric::bindings::traits::vector_stride(v)),
	     size(boost::numeric::bindings::traits::vector_size(v))
	  {}
	};

	template<typename SC,typename INDEX_T>
	struct vector_traits<typical_vector_bridge<SC,INDEX_T> >{
	private:
	  typedef typical_vector_bridge<SC,INDEX_T> V;
	  typedef typename detail::non_const_type_resolve<V>::type V_non_const;
	public:
	  typedef typename V::value_type value_type;
	  typedef typename V::pointer pointer;
	  static typename V::const_pointer const storage(V_non_const const &v){
	    return v.storage_beg;
	  }
	  static pointer const storage(V_non_const &v) {
	    return v.storage_beg;
	  }
	  static INDEX_T const size(V const &v) {
	    return v.size;
	  }
	  static INDEX_T const stride(V const &v) {
	    return v.stride;
	  }
	};

	template<typename SC,typename INDEX_T>
	struct vector_traits<const typical_const_vector_bridge<SC,INDEX_T> >{
	private:
	  typedef const typical_const_vector_bridge<SC,INDEX_T> V;
	  typedef typename detail::non_const_type_resolve<V>::type V_non_const;
	public:
	  typedef typename V::value_type value_type;
	  typedef typename V::pointer pointer;
	  static typename V::const_pointer const storage(V_non_const const &v){
	    return v.storage_beg;
	  }
	  static pointer const storage(V_non_const &v) {
	    return v.storage_beg;
	  }
	  static INDEX_T const size(V const &v) {
	    return v.size;
	  }
	  static INDEX_T const stride(V const &v) {
	    return v.stride;
	  }
	};
	
	template<typename SC,typename INDEX_T,
		 typename MATRIX_STR,typename ORD_TYPE,typename STOR_FORM,
		 unsigned IBase>
	struct typical_sparse_matrix_bridge{
	  typedef SC value_type;
	  typedef const SC *const_value_pointer;
	  typedef SC *value_pointer;
	  typedef const INDEX_T *const_index_pointer;
	  typedef INDEX_T *index_pointer;
	  typedef MATRIX_STR matrix_structure;
	  typedef ORD_TYPE ordering_type;
	  typedef STOR_FORM storage_format;

	  index_pointer const maj_indexs_beg,min_indexs_beg;
	  value_pointer const val_storage_beg;
	  INDEX_T const size1,size2;
	  INDEX_T const num_nonzeros;

	  enum {index_base = IBase};

	  typical_sparse_matrix_bridge(index_pointer const maj_indexs_begi,
				       index_pointer const min_indexs_begi,
				       value_pointer const val_storage_begi,
				       INDEX_T const size1i,
				       INDEX_T const size2i,
				       INDEX_T const nnz)
	    :maj_indexs_beg(maj_indexs_begi),min_indexs_beg(min_indexs_begi),
	     val_storage_beg(val_storage_begi),
	     size1(size1i),size2(size2i),num_nonzeros(nnz)
	  {}
	  template<typename MATRIX_T> explicit
	  typical_sparse_matrix_bridge(MATRIX_T &mi)
	    :maj_indexs_beg(boost::numeric::bindings::traits::
			    sparse_matrix_traits<MATRIX_T>::
			    index1_storage(mi)),
	     min_indexs_beg(boost::numeric::bindings::traits::
			    sparse_matrix_traits<MATRIX_T>::
			    index2_storage(mi)),
	     val_storage_beg(boost::numeric::bindings::traits::
			     sparse_matrix_traits<MATRIX_T>::
			     value_storage(mi)),
	     size1(boost::numeric::bindings::traits::
		   sparse_matrix_traits<MATRIX_T>::size1(mi)),
	     size2(boost::numeric::bindings::traits::
		   sparse_matrix_traits<MATRIX_T>::size2(mi)),
	     num_nonzeros(boost::numeric::bindings::traits::
			  sparse_matrix_traits<MATRIX_T>::num_nonzeros(mi))
	  {}
	};
	template<typename SC,typename INDEX_T,typename MATRIX_STR,
		 typename ORD_TYPE,typename STOR_FORM,unsigned IBase>
	struct
	sparse_matrix_traits<const typical_sparse_matrix_bridge<SC,INDEX_T,
								MATRIX_STR,
								ORD_TYPE,
								STOR_FORM,
								IBase> >{
	private:
	  typedef const typical_sparse_matrix_bridge<SC,INDEX_T,MATRIX_STR,
						     ORD_TYPE,STOR_FORM,IBase>
	  M;
	  typedef typename detail::non_const_type_resolve<M>::type M_non_const;
	public:
	  typedef typename M::value_type value_type;
	  typedef typename M::value_pointer value_pointer;
	  typedef typename M::const_index_pointer const_index_pointer;
	  typedef typename M::index_pointer index_pointer;
	  typedef typename M::matrix_structure matrix_structure;
	  typedef typename M::ordering_type ordering_type;
	  typedef typename M::storage_format storage_format;
	  enum { index_base = M::index_base };

	  static index_pointer const index1_storage(M_non_const &m){
	    return m.maj_indexs_beg;
	  }
	  static typename M::const_index_pointer const
	  index1_storage(M_non_const const &m){
	    return m.maj_indexs_beg;
	  }

	  static index_pointer const index2_storage(M_non_const &m){
	    return m.min_indexs_beg;
	  }
	  static typename M::const_index_pointer const
	  index2_storage(M_non_const const &m){
	    return m.min_indexs_beg;
	  }

	  static value_pointer const value_storage(M_non_const &m){
	    return m.val_storage_beg;
	  }
	  static typename M::const_value_pointer const
	  value_storage(M_non_const const &m){
	    return m.val_storage_beg;
	  }

	  static INDEX_T const size1(M const &m){
	    return m.size1;
	  }
	  static INDEX_T const size2(M const &m){
	    return m.size2;
	  }
	  static INDEX_T const num_nonzeros(M const &m){
	    return m.num_nonzeros;
	  }
	};
      }
    }
  }
}
#endif
