// matrices.h -*-c++-*-
#ifndef _matrices_HDR_
#define _matrices_HDR_

#include<cmath>

#include<boost/numeric/ublas/matrix.hpp>

#include"project1_definitions.h"
#include"arithmetic_tools.h"
#include<fstream>

namespace pu{
  namespace ublas = boost::numeric::ublas ;
  template<typename S> class sparse_matrix_ro;

  template<typename SC> realsc const norm2(SC const & x){return std::abs(x);}

  template<typename SC> class dense_matrix{
  private:
    typedef ublas::matrix<SC,ublas::column_major> m_t;
    m_t m;
    explicit dense_matrix(m_t const & um):m(um) {}
  public:
    typedef typename m_t::reference reference;
    typedef typename m_t::const_reference const_reference;
    dense_matrix():m() {}
    dense_matrix(size_t ROWS,size_t COLS) :m(ROWS,COLS) {}
    template<typename SC2> dense_matrix(sparse_matrix_ro<SC2> const &sm_ro);
    void resize(size_t ROWS,size_t COLS);
    size_t const size1() const { return m.size1(); }
    size_t const size2() const { return m.size2(); }
    const_reference const operator()(size_t const i,size_t const j)const
    { return m(i,j);}
    reference operator()(size_t const i,size_t const j)
    { return m(i,j);}
    realsc const norm2() const ;

    template<typename S_OTHER>
    dense_matrix<typename add_promotion_traits<SC,S_OTHER>::promote_type> const
    operator-(dense_matrix<S_OTHER> const & m2) const {
      typedef dense_matrix<typename add_promotion_traits<SC,S_OTHER>::promote_type> result_type;
      return result_type(m-m2.m);
    }

    std::ostream & print(std::ostream & c) const ;
    realsc const cond() const ;
    dense_matrix const & inv_self();
    template<typename VT> VT const solve(VT const &rhs)const;

    m_t const & inner_ublas() const { return m; }
    m_t & inner_ublas() { return m; }

    typedef typename m_t::value_type value_type;
    typedef typename m_t::const_iterator2 const_iterator2;
    const_iterator2 const begin2() const { return m.begin2(); }
    const_iterator2 const end2() const { return m.end2(); }

    typedef typename m_t::const_iterator1 const_iterator1;
    const_iterator1 const begin1() const { return m.begin1(); }
    const_iterator1 const end1() const { return m.end1(); }

    void clear() { m.clear(); }

    template<typename VT> void
    setcolumn(size_t const ncol,VT const &vin){
      //for(size_t nr(0);nr<vin.size();++nr){ operator()(nr,nc)=vin[nr];
      //}
      assert(size1()==vin.size());
      ublas::matrix_column<m_t> col(m,ncol);
      copy(vin.begin(),vin.end(),col.begin());
    }
    ublas::matrix_column<const m_t> const getcolumn(size_t const ncol) const {
      //can copy to and return a std::vector or pu::ace instead.
      return  ublas::matrix_column<const m_t> (m,ncol);
    }

    typedef typename m_t::iterator2 iterator2;
    iterator2 const begin2() { return m.begin2(); }
    iterator2 const end2() { return m.end2(); }

    typedef typename m_t::iterator1 iterator1;
    iterator1 const begin1() { return m.begin1(); }
    iterator1 const end1() { return m.end1(); }

    typedef ublas::matrix_slice<const m_t> const_matrix_slice_t;
    const_matrix_slice_t const matrix_slice(size_t const sta_r,
					    size_t const step_r,
					    size_t const n_r,
					    size_t const sta_c,
					    size_t const step_c,
					    size_t const n_c) const {
      return const_matrix_slice_t(m,ublas::slice(sta_r,step_r,n_r),
				  ublas::slice(sta_c,step_c,n_c));
    }
    explicit dense_matrix(const_matrix_slice_t const &mslice_in)
      :m(mslice_in) {}
  };

  //no alias !
  template<typename MT,typename V,typename U> void
  mvmult_A_v(MT const & A,V const &v,U & u);
  template<typename MT,typename V,typename U> void
  mvmult_At_v(MT const & A,V const &v,U & u);
  template<typename MTA,typename MTB,typename MTC> void
  mxmult_A_B(MTA const &ma,MTB const &mb,MTC &mc);
  template<typename MTA,typename MTB,typename MTC> void
  mxmult_A_Bt(MTA const &ma,MTB const &mb,MTC &mc);

}//of pu

#include"project1_definitions.h"
namespace pu{
  template<typename MT> void ReadDenseMatrix(std::string fname,MT &m){
    std::ifstream C(fname.c_str());
    ReadDenseMatrix(C,m);
  }
  template<typename MT>
  std::istream &ReadDenseMatrix(std::istream & C,MT &m);
  template<typename S> std::istream &
  operator>>(std::istream & C,pu::dense_matrix<S> &m){
    return ReadDenseMatrix(C,m);
  }
  template<typename MT> void WriteDenseMatrix(std::string fname,MT const &m){
    std::ofstream C(fname.c_str());
    WriteDenseMatrix(C,m);
  }
  template<typename MT>
  std::ostream & WriteDenseMatrix(std::ostream &C,MT const &m);

} // of pu

#include"sparse_matrices.h"
namespace pu{
  template<typename S> template<typename SC2>
  dense_matrix<S>::dense_matrix(sparse_matrix_ro<SC2> const &sm_ro)
    :m(sm_ro.ublas_matrix()) {}
}//of pu

#include"matrices.cct"

#endif
