//sparse_matrices.cc -*-c++-*-
#include"sparse_matrices.h"

#include <boost/numeric/bindings/umfpack/umfpack.hpp>

#include<complex>

#include<addins_for_num_bindings.hpp>

namespace sparse_matrices_pr{
  namespace umf = boost::numeric::bindings::umfpack;

  template<typename S> class umf_direct_sol_data_impl{
  public:
    typedef ublas::compressed_matrix<S,ublas::column_major,0,
				     ublas::unbounded_array<int>,
				     ublas::unbounded_array<S> > umf_spro_t;
    typedef umf::numeric_type<S> Numeric_t;
    typedef umf::symbolic_type<S> Symbolic_t;

    typedef boost::numeric::bindings::traits::row_major_t row_major_t;
    typedef boost::numeric::bindings::traits::column_major_t column_major_t;
    template<typename SPARSE_MAT_T> static umf_spro_t const
    mcopy(SPARSE_MAT_T &mi,row_major_t const &);
    template<typename SPARSE_MAT_T> static umf_spro_t const
    mcopy(SPARSE_MAT_T &mi,column_major_t const &);
    template<typename SPARSE_MAT_T> static umf_spro_t const
    mcopy(SPARSE_MAT_T &mi){
      namespace traits = boost::numeric::bindings::traits;
      return mcopy(mi,typename traits::sparse_matrix_traits<SPARSE_MAT_T>::
		   ordering_type());
    }

    Numeric_t Numeric; //non-copyable
    void init_Numeric(umf_spro_t const &m);
  public:
    umf_direct_sol_data_impl(umf_spro_t const &m):Numeric() {
      init_Numeric(m);
    }
    template<typename VS_T,typename VT_T> void
    mldivide(int const sys,umf_spro_t const &m,
	     VS_T const & vs,VT_T & vt) const ;
    template<typename VS_T,typename VT_T> void
    scale(VS_T const & vs,VT_T & vt) const ;
  };

  template<typename S>
  class umf_factorized_impl{
    BOOST_STATIC_ASSERT((boost::mpl::or_<boost::is_same<S,double>,
			 boost::is_same<S,std::complex<double> > >::value));
    typedef umf_direct_sol_data_impl<S> impl_t;
    typedef typename impl_t::umf_spro_t umf_spro_t;

    umf_spro_t const mat;
    impl_t me; //makes noncopiable

    static int const SysK2U(SysKind const k);

  public:
    template<typename MATRIX_T> explicit
    umf_factorized_impl(MATRIX_T const &mi)
      :mat(impl_t::mcopy(mi)),
       me(mat)
    {}
    template<typename VECTOR_ST,typename VECTOR_TT>
    void mldivide(int const sys,
		  VECTOR_ST const &vs,VECTOR_TT &vt) const {
      me.mldivide(sys,mat,vs,vt);
    }
    template<typename VECTOR_ST,typename VECTOR_TT>
    void scale(VECTOR_ST const &vs,VECTOR_TT &vt) const {
      me.scale(vs,vt);
    }
  };

  template<typename S> void
  umf_direct_sol_data_impl<S>::
  init_Numeric(umf_spro_t const &m){
    umf::control_type<S> Control;
    umf::info_type<S> Info;
    Control.defaults();
    Symbolic_t Symbolic;  umf::symbolic(m,Symbolic);
    int status = umf::numeric(m,Symbolic,Numeric,Control,Info);
    umf::report_control(Control); // no output by default
    umf::report_status(Control,status); // may issue warning - if singular
    umf::report_info(Control,Info); // no output by default
  }

  template<typename S> template<typename VS_T,typename VT_T> void
  umf_direct_sol_data_impl<S>::
  mldivide(int const sys,umf_spro_t const &m,
	   VS_T const & vs,VT_T & vt)const{
    umf::control_type<S> Control;
    umf::info_type<S> Info;
    Control.defaults();
    int const status=umf::solve(sys,m,vt,vs,Numeric,
				Control,Info);
    umf::report_control(Control); // no output by default
    umf::report_status(Control,status); // may issue warning - if singular
    umf::report_info(Control,Info); // no output by default
  }





  template<typename S> template<typename VS_T,typename VT_T> void
  umf_direct_sol_data_impl<S>::
  scale(VS_T const & vs,VT_T & vt)const{
    umf::control_type<S> Control;
    umf::info_type<S> Info;
    Control.defaults();
    int const status=umf::scale(vt,vs,Numeric);
    umf::report_control(Control); // no output by default
    umf::report_status(Control,status); // may issue warning - if singular
    umf::report_info(Control,Info); // no output by default
  }




  template<typename S> template<typename SPARSE_MAT_T>
  class umf_direct_sol_data_impl<S>::umf_spro_t const
  umf_direct_sol_data_impl<S>::
  mcopy(SPARSE_MAT_T &mi,column_major_t const &){
    namespace traits = boost::numeric::bindings::traits;
    typedef traits::sparse_matrix_traits<SPARSE_MAT_T> smtraits;
    umf_spro_t result(smtraits::size1(mi),smtraits::size2(mi),
		      smtraits::num_nonzeros(mi));
    //column major order of passing
    for(unsigned int j(0);j<smtraits::size2(mi);++j){
      typename smtraits::index_pointer const curr_col_ix_end
	( smtraits::index2_storage(mi) +
	  (smtraits::index1_storage(mi)[j+1] - smtraits::index_base));
      typename smtraits::value_pointer v_p
	( smtraits::value_storage(mi) +
	  (smtraits::index1_storage(mi)[j] - smtraits::index_base));
      for(typename smtraits::index_pointer r_p
	    ( smtraits::index2_storage(mi) +
	      (smtraits::index1_storage(mi)[j] - smtraits::index_base));
	  r_p<curr_col_ix_end;++r_p,++v_p){
	result.push_back((*r_p - smtraits::index_base),j,*v_p);
      }
    }
    return result;
  }

  template<typename S> template<typename SPARSE_MAT_T>
  class umf_direct_sol_data_impl<S>::umf_spro_t const
  umf_direct_sol_data_impl<S>::
  mcopy(SPARSE_MAT_T &mi,row_major_t const &){
    namespace traits = boost::numeric::bindings::traits;
    typedef traits::sparse_matrix_traits<SPARSE_MAT_T> smtraits;
    ublas::compressed_matrix<S,ublas::row_major,0,
      ublas::unbounded_array<int>,
      ublas::unbounded_array<S> >
      result(smtraits::size1(mi),smtraits::size2(mi),
	     smtraits::num_nonzeros(mi));
    //row major order of passing
    for(unsigned i(0);i<smtraits::size1(mi);++i){
      typename smtraits::index_pointer const curr_row_ix_end
	( smtraits::index2_storage(mi) +
	  (smtraits::index1_storage(mi)[i+1] - smtraits::index_base));
      typename smtraits::value_pointer v_p
	( smtraits::value_storage(mi) +
	  (smtraits::index1_storage(mi)[i] - smtraits::index_base) );
      for(typename smtraits::index_pointer
	    c_p( smtraits::index2_storage(mi) +
		 (smtraits::index1_storage(mi)[i] - smtraits::index_base));
	  c_p<curr_row_ix_end;++c_p,++v_p){
	result.push_back(i,(*c_p - smtraits::index_base),*v_p);
      }
    }
    return umf_spro_t(result);
  }

  template<typename S>
  umf_direct_sol_data<S>::
  umf_direct_sol_data(ublas_spro_t const &m)
    :me(new impl_t(impl_t::mcopy(m)))
  {}

  template<typename S> void umf_direct_sol_data<S>::
  mldivide(SysKind const sys,ublas_spro_t const &m,
	   boost::numeric::bindings::traits::
	   typical_const_vector_bridge<S> const &vs,
	   boost::numeric::bindings::traits::
	   typical_vector_bridge<S> &vt) const{
    me->mldivide(SysK2U(sys),impl_t::mcopy(m),vs,vt);
  }

  template<typename S>
  int const umf_direct_sol_data<S>::
  SysK2U(SysKind const k){
    int result;
    switch(k){
    case SP_A:
      result = UMFPACK_A;
      break;
    case SP_At: 
      result = UMFPACK_At;
      break;
    default: std::exit(-1);
    }
    return result;
  }

  //explicit instantiations
  template class umf_direct_sol_data<double>;
  template class umf_direct_sol_data<std::complex<double> >;

}//of sparse_matrices_pr

namespace pu{

  template<typename S>
  template<typename S1,typename IND,typename MSTR,
	   typename ORD,typename STF,unsigned IBase>
  void umf_factorized<S>::
  init(boost::numeric::bindings::traits::
       typical_sparse_matrix_bridge<S1,IND,MSTR,ORD,STF,IBase> const &m){
    me.reset(new impl_t(m));
  }

  template<typename S> void umf_factorized<S>::
  mldivide(int const sys,
	   boost::numeric::bindings::traits::
	   typical_const_vector_bridge<S> const &vs,
	   boost::numeric::bindings::traits::
	   typical_vector_bridge<S> &vt) const{
    me->mldivide(sys,vs,vt);
  }
  template<typename S> void umf_factorized<S>::
  scale(boost::numeric::bindings::traits::
	typical_const_vector_bridge<S> const &vs,
	boost::numeric::bindings::traits::
	typical_vector_bridge<S> &vt) const{
    me->scale(vs,vt);
  }

  //explicit instantiations

  template void umf_factorized<double>::
  init(boost::numeric::bindings::traits::
       typical_sparse_matrix_bridge
       <const double,
       const std::size_t,
       boost::numeric::bindings::traits::general_t,
       boost::numeric::bindings::traits::column_major_t,
       boost::numeric::bindings::traits::compressed_t,
       0> const &);
  template void umf_factorized<std::complex<double> >::
  init(boost::numeric::bindings::traits::
       typical_sparse_matrix_bridge
       <const std::complex<double>,
       const std::size_t,
       boost::numeric::bindings::traits::general_t,
       boost::numeric::bindings::traits::column_major_t,
       boost::numeric::bindings::traits::compressed_t,
       0> const &);

  // Non templates, so ambiguous.
  //   void umf_factorized<double>::
  //   mldivide(int const sys,
  //   	   boost::numeric::bindings::traits::
  //   	   typical_const_vector_bridge<double> const &vs,
  //   	   boost::numeric::bindings::traits::
  //   	   typical_vector_bridge<double> &vt) const;
  //   template void umf_factorized<std::complex<double> >::
  //   mldivide(int const sys,
  // 	   boost::numeric::bindings::traits::
  // 	   typical_const_vector_bridge<std::complex<double> > const &vs,
  // 	   boost::numeric::bindings::traits::
  // 	   typical_vector_bridge<std::complex<double> > &vt) const;
  // So forced:
  template class umf_factorized<double>;
  template class umf_factorized<std::complex<double> >;

}//of pu
