// sparse_matrices.h -*-c++-*-
#ifndef _sparse_matrices_HDR_
#define _sparse_matrices_HDR_

#include<boost/mpl/if.hpp>
#include<boost/mpl/or.hpp>
#include<boost/static_assert.hpp>
#include<boost/type_traits/is_same.hpp>

#include <boost/shared_ptr.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/bindings/traits/ublas_sparse.hpp>

#include <boost/numeric/bindings/traits/std_vector.hpp>
#include <boost/numeric/bindings/traits/ublas_vector.hpp>

#include<iostream>
#include<map>
#include<vector>
#include<complex>

#include<addins_for_num_bindings.hpp>
#include"pr1arr.h" // for "ace"


namespace sparse_matrices_pr{
  enum SysKind {SP_A,SP_At};
  namespace ublas = boost::numeric::ublas;
  template<typename S> class umf_direct_sol_data_impl; //forward
  template<typename S>
  class umf_direct_sol_data{
    BOOST_STATIC_ASSERT((boost::mpl::or_<boost::is_same<S,double>,
			 boost::is_same<S,std::complex<double> > >::value));
    typedef umf_direct_sol_data_impl<S> impl_t;
    typedef ublas::compressed_matrix<S,ublas::column_major,0,
				     ublas::unbounded_array<size_t>,
				     ublas::unbounded_array<S> >
    ublas_spro_t;

    boost::shared_ptr<impl_t> me;

    static int const SysK2U(SysKind const k);

  public:
    explicit umf_direct_sol_data(ublas_spro_t const &m);
    void mldivide(SysKind const sys,ublas_spro_t const &m,
		  boost::numeric::bindings::traits::
		  typical_const_vector_bridge<S> const &vs,
		  boost::numeric::bindings::traits::
		  typical_vector_bridge<S> &vt) const;
    template<typename VS_T,typename VT_T>
    void mldivide(SysKind const sys,ublas_spro_t const &m,
		  VS_T const &vs,VT_T &vt) const{
      namespace traits = boost::numeric::bindings::traits;
      traits::typical_vector_bridge<S> bvt(vt);
      mldivide(sys,m,traits::typical_const_vector_bridge<S>(vs),bvt);
    }
  };
}//of sparse_matrices_pr
namespace pu{
  using sparse_matrices_pr::SysKind;
  using sparse_matrices_pr::SP_A;
  using sparse_matrices_pr::SP_At;

  template<typename S> class sparse_matrix_ro;

  template<typename S> class sparse_matrix_rw{
    // This have bad performance because of insertions into array
    // which caused reallocations of O(N) per insertion.
    //     typedef sparse_matrices_pr::ublas::
    //     mapped_matrix<S,sparse_matrices_pr::ublas::column_major,
    // 		  sparse_matrices_pr::ublas::map_array<std::size_t,S> >
    //     ublas_sprw_t;
    typedef sparse_matrices_pr::ublas::
    mapped_matrix<S,sparse_matrices_pr::ublas::column_major,
		  sparse_matrices_pr::ublas::map_std<std::size_t,S> >
    ublas_sprw_t;
    ublas_sprw_t m;

  public:
    typedef S value_type;
    typedef typename ublas_sprw_t::reference reference;
    typedef typename ublas_sprw_t::const_reference const_reference;
    ublas_sprw_t const & ublas_matrix() const { return m; } //inner interface
    ublas_sprw_t & ublas_matrix() { return m; } //inner interface
    sparse_matrix_rw(unsigned const sz1,unsigned const sz2)
      :m(sz1,sz2) {}
    sparse_matrix_rw(unsigned const sz1,unsigned const sz2,unsigned const nnzi)
      :m(sz1,sz2,nnzi) {}
    template<typename S_OTHER>
    explicit sparse_matrix_rw(sparse_matrix_ro<S_OTHER> const &m_in);
    void push_back(size_t const i,size_t const j,S const &s_in){
      //!? inner hint-iterator
      m.insert(i,j,s_in);
    }
    size_t const size1() const { return m.size1(); }
    size_t const size2() const { return m.size2(); }
    const_reference const operator()(unsigned const i,unsigned const j) const
    {return m(i,j);}
    reference operator()(unsigned const i,unsigned const j){return m(i,j);}

    void clear() { m.clear(); }
    sparse_matrix_rw const & operator+=(sparse_matrix_ro<S> const &so);
    sparse_matrix_rw const & operator-=(sparse_matrix_ro<S> const &so);
    template<typename S_OTHER>
    sparse_matrix_rw const & operator*=(S_OTHER const &s);

    typedef typename ublas_sprw_t::iterator2 iterator2;
    iterator2 const begin2() { return m.begin2(); }
    iterator2 const end2() { return m.end2(); }

    typedef typename ublas_sprw_t::iterator1 iterator1;
    iterator1 const begin1() { return m.begin1(); }
    iterator1 const end1() { return m.end1(); }

    std::ostream & print(std::ostream & c) const
    {return c<<m;}

    void erase(unsigned const i,unsigned const j){
      m.erase(i,j);
    }
    void insert(unsigned const i,unsigned const j,S const &s){
      m.insert(i,j,s);
    }


    //for nnz and let them copy it
    typedef typename ublas_sprw_t::const_iterator2 const_iterator2;
    const_iterator2 const begin2() const { return m.begin2(); }
    const_iterator2 const end2() const { return m.end2(); }

    typedef typename ublas_sprw_t::const_iterator1 const_iterator1;
    const_iterator1 const begin1() const { return m.begin1(); }
    const_iterator1 const end1() const { return m.end1(); }

    size_t const nnz() const {
      size_t count(0);
      //column major order of passing
      for(const_iterator2 it2(begin2());it2 != end2(); ++it2){
	for(const_iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  ++count;
	}
      }
      return count;
    }

  };

  template<typename S,typename IT> void
  ZeroMRows(sparse_matrix_rw<S> &KS,IT const & beg,IT const &end);

  template<typename S> class sparse_matrix_ro{
    typedef sparse_matrices_pr::ublas::
    compressed_matrix<S,sparse_matrices_pr::ublas::column_major,0,
		      sparse_matrices_pr::ublas::unbounded_array<size_t>,
		      sparse_matrices_pr::ublas::unbounded_array<S> >
    ublas_spro_t;
    ublas_spro_t const m;
    //keep it 'const' this way,because in ublas operations with
    //compressed matrices are extremally unefficient ( O(nnz^2) )

    static ublas_spro_t const mcopy(sparse_matrix_rw<S> const &mi){
      typedef typename sparse_matrix_rw<S>::const_iterator1 ITT1;
      typedef typename sparse_matrix_rw<S>::const_iterator2 ITT2;
      ublas_spro_t result(mi.size1(),mi.size2(),mi.nnz());
      //column major order of passing
      for(ITT2 it2(mi.begin2());it2 != mi.end2(); ++it2){
	for(ITT1 it1(it2.begin());it1!=it2.end(); ++it1){
	  result.insert(it1.index1(),it1.index2(),*it1);
	}
      }
      return result;
    }
  public:
    typedef S value_type;
    typedef typename ublas_spro_t::reference reference;
    typedef typename ublas_spro_t::const_reference const_reference;

    ublas_spro_t const & ublas_matrix() const { return m; }
    typedef typename sparse_matrices_pr::umf_direct_sol_data<S> LinSolData_t;
    explicit sparse_matrix_ro(sparse_matrix_rw<S> const &mi):m(mcopy(mi)) {}

    size_t const size1() const { return m.size1(); }
    size_t const size2() const { return m.size2(); }

    LinSolData_t const LinSolData() const { return LinSolData_t(m); }
    const_reference const operator()(unsigned const i,unsigned const j) const {
      return m(i,j);
    }
    template<typename VS_T,typename VT_T>
    void mldivide(SysKind const sys,LinSolData_t const &L,
		  VS_T const &vs,VT_T &vt) const {
      L.mldivide(sys,m,vs,vt);
    }

    //How do you encapsulate expression templates ?
  private:
    explicit sparse_matrix_ro(ublas_spro_t const &m_in):m(m_in) {}
  public:
    size_t const nnz() const {
      size_t count(0);
      //column major order of passing
      for(const_iterator2 it2(begin2());it2 != end2(); ++it2){
	for(const_iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  ++count;
	}
      }
      return count;
    }

  public:
    sparse_matrix_ro const operator-()const {
      ublas_spro_t preresult(m);
      typedef typename ublas_spro_t::iterator1 iterator1;
      typedef typename ublas_spro_t::iterator2 iterator2;
      for(iterator2 it2(preresult.begin2());it2!=preresult.end2();++it2){
	for(iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  *it1 = -(*it1);
	}
      }
      return sparse_matrix_ro(preresult);  //by copying
    }
    sparse_matrix_ro const transposed() const {
      return sparse_matrix_ro(sparse_matrices_pr::ublas::trans(m));
    }
    static sparse_matrix_ro const eye(size_t const ni){
      return sparse_matrix_ro
	(sparse_matrices_pr::ublas::identity_matrix<value_type>(ni));
    }

    typedef typename ublas_spro_t::const_iterator2 const_iterator2;
    const_iterator2 const begin2() const { return m.begin2(); }
    const_iterator2 const end2() const { return m.end2(); }

    typedef typename ublas_spro_t::const_iterator1 const_iterator1;
    const_iterator1 const begin1() const { return m.begin1(); }
    const_iterator1 const end1() const { return m.end1(); }

    std::ostream & print(std::ostream & c) const
    {return c<<m;}
  };

  template<typename T> sparse_matrix_ro<T> const
  concatv(sparse_matrix_ro<T> const &m1,sparse_matrix_ro<T> const &m2);
  template<typename T> sparse_matrix_ro<T> const
  concath(sparse_matrix_ro<T> const &m1,sparse_matrix_ro<T> const &m2);

  template<typename S> template<typename S_OTHER>
  sparse_matrix_rw<S>::sparse_matrix_rw(sparse_matrix_ro<S_OTHER> const &m_in)
    :m(m_in.ublas_matrix()) {}

  template<typename S> std::ostream &
  operator<<(std::ostream &c,sparse_matrix_rw<S> const &sm)
  {return sm.print(c);}
  template<typename S> std::ostream &
  operator<<(std::ostream &c,sparse_matrix_ro<S> const &sm)
  {return sm.print(c);}

  template<typename S> inline sparse_matrix_rw<S> const &
  sparse_matrix_rw<S>::operator+=(sparse_matrix_ro<S> const &so)
  {m+=so.ublas_matrix();  return *this;}
  template<typename S> inline sparse_matrix_rw<S> const &
  sparse_matrix_rw<S>::operator-=(sparse_matrix_ro<S> const &so)
  {m-=so.ublas_matrix();  return *this;}

  template<typename S> template<typename S_OTHER>
  sparse_matrix_rw<S> const & sparse_matrix_rw<S>::
  operator*=(S_OTHER const &s){
    for(iterator2 it2(begin2());it2!=end2();++it2){
      for(iterator1 it1(it2.begin());it1!=it2.end();++it1){
	(*it1) *= s;
      }
    }
    return *this;
  }

  template<typename S,typename VS_T,typename VT_T>
  void mldivide(typename sparse_matrix_ro<S>::LinSolData_t const &L,
		sparse_matrix_ro<S> const &M,VS_T const &vs,VT_T &vt){
    mldivide(SP_A,L,M,vs,vt);
  }

  template<typename S,typename VS_T,typename VT_T>
  void mldivide(SysKind const sys,
		typename sparse_matrix_ro<S>::LinSolData_t const &L,
		sparse_matrix_ro<S> const &M,VS_T const &vs,VT_T &vt){
    M.mldivide(sys,L,vs,vt);
  }
}//of pu

namespace sparse_matrices_pr{
  template<typename S> class umf_factorized_impl; //forward
}// of sparse_matrices_pr

namespace pu{

  template<typename S>
  class umf_factorized{
    typedef sparse_matrices_pr::umf_factorized_impl<S> impl_t;
    boost::shared_ptr<impl_t> me;

    //   just because ctor can't call other ctor, and it is needed to
    // encapsulate the use of "impl_t".
    template<typename S1,typename IND,typename MSTR,
	     typename ORD,typename STF,unsigned IBase>
    void init(boost::numeric::bindings::traits::
	      typical_sparse_matrix_bridge<S1,IND,MSTR,ORD,STF,IBase>
	      const &m);

    template<typename T> struct resolve_ix_t;
    template<typename T> struct resolve_ix_t<T *>{
      typedef T type;
    };
    template<typename MT> struct resolve_m_bridge{
      typedef boost::numeric::bindings::traits::sparse_matrix_traits<MT>
      MTraits;
      typedef boost::numeric::bindings::traits::typical_sparse_matrix_bridge
      <typename MTraits::value_type const,
       typename resolve_ix_t<typename MTraits::index_pointer>::type const,
       typename MTraits::matrix_structure,
       typename MTraits::ordering_type,
       typename MTraits::storage_format,
       MTraits::index_base> type;
    };

  public:
    template<typename MATRIX_T> explicit
    umf_factorized(MATRIX_T const &mi)
    {
      init(typename resolve_m_bridge<MATRIX_T>::type(mi));
    }
    template<typename S1,typename Orient_t> explicit
    umf_factorized(boost::numeric::ublas::compressed_matrix<S1,Orient_t>
		   const &mi)
    {
      // consider non-const (well documented "why") "mi" parameter
      //mi.complete_index1_data();
      init(typename resolve_m_bridge
	   <boost::numeric::ublas::compressed_matrix<S1,Orient_t> >::type(mi));
    }
    void mldivide(int const sys,
		  boost::numeric::bindings::traits::
		  typical_const_vector_bridge<S> const &vs,
		  boost::numeric::bindings::traits::
		  typical_vector_bridge<S> &vt) const;
    template<typename VECTOR_ST,typename VECTOR_TT> void
    mldivide(int const sys,VECTOR_ST const &vs,VECTOR_TT &vt) const{
      namespace traits = boost::numeric::bindings::traits;
      traits::typical_vector_bridge<S> bvt(vt);
      mldivide(sys,traits::typical_const_vector_bridge<S>(vs),bvt);
    }

    void scale(boost::numeric::bindings::traits::
	       typical_const_vector_bridge<S> const &vs,
	       boost::numeric::bindings::traits::
	       typical_vector_bridge<S> &vt) const;
    template<typename VECTOR_ST,typename VECTOR_TT> void
    scale(VECTOR_ST const &vs,VECTOR_TT &vt) const{
      namespace traits = boost::numeric::bindings::traits;
      traits::typical_vector_bridge<S> bvt(vt);
      scale(traits::typical_const_vector_bridge<S>(vs),bvt);
    }
  };



}//of namespace pu

#include"sparse_matrices.cct"
#endif
