// telescopic_eigen0.cc -*-c++-*-

#include<base/parameter_handler.h>

#include<iostream>
#include<string>

#include<GlobalTimer.h>
#include<basic_cmdline_params_reader.h>
#include<fem2prj_run.h>
#include<run_wrapper.h>

namespace pu{
  //forward
  template<typename SC> class LAL_TypeEigDriverNonSym;
  template<> class LAL_TypeEigDriverNonSym<realsc>;

  template<int dim> class run_eigen0 {
    Fem2PrjRun<dim> f2;

    void
    solve_dump_eigs
    (boost::shared_ptr<LAL_TypeEigDriverNonSym<realsc> > const &p,
     realsc const tol,std::string const &which,unsigned const nev,
     std::string const &fname_suffix) const;
  public:
    run_eigen0(bool const IsConfirmParams,unsigned const RunNo,
	       ParameterHandler &prm)
      :f2(IsConfirmParams,RunNo,prm)
    {}

    static void declare_parameters (ParameterHandler &prm){
      Fem2PrjRun<dim>::declare_parameters(prm);
    }
    void run();
    virtual ~run_eigen0() {}
  };

}//of pu

int main(int const argc,char const * const argv[]){

  using namespace pu;
  using namespace std;
  cout << "Time: "<<GlobalTimer.elapsed()<<"\n";
  basic_cmdline_params_reader_t reader_cmdline(argc,argv);
  reader_cmdline.require_min(1).require_max(3);

  int dim;
  reader_cmdline.read_next_arg(dim,"dim");
  string input_fname;
  reader_cmdline.read_next_arg(input_fname,"input file name");
  exit_check_that(dim==2 || dim==3,"dim==2 || dim==3");

  //
  //optional args
  //
  bool IsConfirmParams(true);
  if(reader_cmdline.has_additional_args()){
    reader_cmdline.read_next_arg(IsConfirmParams,
				 "IsConfirmParams");
  }

  reader_cmdline.require_terminated();

  class MultipleParameterLoop prm;
  switch (dim) {
  case 2:
    {
      RunWrapper<run_eigen0<2> > problem(IsConfirmParams);
      problem.declare_parameters(prm);
      prm.read_input(input_fname);
      prm.loop(problem);
    }
    break;
  case 3:
    {
      RunWrapper<run_eigen0<3> > problem(IsConfirmParams);
      problem.declare_parameters(prm);
      prm.read_input(input_fname);
      prm.loop(problem);
    }
    break;
  }

  return 0;
}

#include<lac/block_sparse_matrix.h>

#include<complex>
#include<vector>
#include<fstream>

#include<boost/shared_ptr.hpp>
#include<boost/numeric/ublas/matrix_sparse.hpp>

#include<spatial_fem_brick.interface.h>
#include<ublasdealii.h>
#include<eigen_solvers_wrappers.h>

namespace pu{
  template<int dim> void run_eigen0<dim>::run(){
    namespace ublas = boost::numeric::ublas;
    boost::shared_ptr<INTERFACE_SpatialFemEHBrick<dim> const>
      pSpatialPart(f2.PrepareSpatialPart());
    boost::shared_ptr<LAL_TypeEigDriverNonSym<realsc> > pEIG_LAL;
    unsigned size_of_problem(0);
    {
      BlockSparsityPattern MP,KP;
      BlockSparseMatrix<realsc> M,K;
      pSpatialPart->ReinitMK(MP,M,KP,K);
      boost::shared_ptr<ublas::compressed_matrix<realsc,ublas::column_major> >
	pM(new ublas::compressed_matrix<realsc,ublas::column_major>()),
	pK(new ublas::compressed_matrix<realsc,ublas::column_major>());
      uBLASDealii::Dealii2uBLAS(M,*pM);
      uBLASDealii::Dealii2uBLAS(K,*pK);
      *pK = - *pK;

      //to make it compatible with standard csr/csc formats
      pM->complete_index1_data();

      pEIG_LAL.reset(new LAL_TypeEigDriverNonSym<realsc>(pM,pK));

      size_of_problem = pK->size1();
    }

    solve_dump_eigs(pEIG_LAL,1e-6,"LM",size_of_problem,"lm_eigs.gnuplot");
  }

  template<int dim> void run_eigen0<dim>::
  solve_dump_eigs
  (boost::shared_ptr<LAL_TypeEigDriverNonSym<realsc> > const &p,
   realsc const tol,std::string const &which,unsigned const nev,
   std::string const &fname_suffix) const {
    p->solve(tol,which,nev);
    std::vector<std::complex<realsc> > eigs;
    p->get_all_converged_eigs(eigs);
    std::string const ofname(f2.prefix_name()+fname_suffix);
    std::ofstream fout(ofname.c_str());
    for(unsigned i(0);i<eigs.size();++i){
      fout << eigs[i].real() << ' ' << eigs[i].imag() << '\n';
    }
  }

}//of pu
