// ublasdealii.h -*-c++-*-
#ifndef _ublasdealii_HDR_
#define _ublasdealii_HDR_

#include<lac/sparse_matrix.h>
#include<lac/block_sparse_matrix.h>

#include"project1_definitions.h"
#include<boost/numeric/ublas/matrix_sparse.hpp>
#include<vector>
#include<cassert>

namespace pu{
  class uBLASDealii {
  private:
    typedef boost::numeric::ublas::column_major column_major;
    typedef boost::numeric::ublas::row_major row_major;

    template<typename SC,typename DM> static void
    helper_d2u(DM const &dm,boost::numeric::ublas::
	       compressed_matrix<SC,row_major> &result){
      //result.resize(dm.m(),dm.n(),/*no such arg:dm.n_nonzero_elements()*/);
      result = boost::numeric::ublas::compressed_matrix<SC,row_major>
	(dm.m(),dm.n(),dm.n_nonzero_elements());

      //   Dealii matrix can have diagonal optimizations, be block based etc,
      // so this naive way doesn't do.
      //for(typename DM::const_iterator ii(dm.begin());ii!=dm.end();++ii){
      //result.push_back(ii->row(),ii->column(),ii->value());
      //}

      boost::numeric::ublas::coordinate_matrix<double,row_major>
	pre_result(dm.m(),dm.n(),dm.n_nonzero_elements());
      for(typename DM::const_iterator ii(dm.begin());ii!=dm.end();++ii){
	pre_result.append_element(ii->row(),ii->column(),ii->value());
      }

      //how do I explicitly say to coordinate_matrix to sort/assemble itself.?

      boost::numeric::ublas::noalias(result) = pre_result;
    }
    template<typename SC,typename DM> static void
    helper_d2u(DM const &dm,boost::numeric::ublas::
	       compressed_matrix<SC,column_major> &result){
      boost::numeric::ublas::compressed_matrix<SC,row_major>
	pre_result;
      helper_d2u(dm,pre_result);

      //result.resize(dm.m(),dm.n(),/*no such arg:dm.n_nonzero_elements()*/);
      result.resize(dm.m(),dm.n(),false);

      boost::numeric::ublas::noalias(result) = pre_result;
    }
  public:

    template<typename SC> static void
    Dealii2uBLAS(BlockSparseMatrix<SC> const &dm,
		 boost::numeric::ublas::
		 compressed_matrix<SC,row_major> &result){
      helper_d2u(dm,result);
    }
    template<typename SC> static void
    Dealii2uBLAS(BlockSparseMatrix<SC> const &dm,
		 boost::numeric::ublas::
		 compressed_matrix<SC,column_major> &result){
      helper_d2u(dm,result);
    }
    template<typename SC> static void
    Dealii2uBLAS(SparseMatrix<SC> const &dm,
		 boost::numeric::ublas::
		 compressed_matrix<SC,row_major> &result){
      helper_d2u(dm,result);
    }
    template<typename SC> static void
    Dealii2uBLAS(SparseMatrix<SC> const &dm,
		 boost::numeric::ublas::
		 compressed_matrix<SC,column_major> &result){
      helper_d2u(dm,result);
    }
    
    template<typename SC> static void
    uBLAS2Dealii(boost::numeric::ublas::
		 compressed_matrix<SC,column_major> const &us,
		 SparsityPattern &ptrn,SparseMatrix<SC> &result){
      std::vector<unsigned> row_lengths(us.size1());
      typedef typename boost::numeric::ublas::
	compressed_matrix<SC,column_major>::const_iterator1 const_iterator1;
      typedef typename boost::numeric::ublas::
	compressed_matrix<SC,column_major>::const_iterator2 const_iterator2;

      for(const_iterator2 it2(us.begin2());it2 != us.end2(); ++it2){
        for(const_iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  row_lengths[it1.index1()]+=1;
	}
      }
      ptrn.reinit(us.size1(),us.size2(),row_lengths);
      for(const_iterator2 it2(us.begin2());it2 != us.end2(); ++it2){
        for(const_iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  ptrn.add(it1.index1(),it1.index2());
	}
      }
      result.reinit(ptrn);
      for(const_iterator2 it2(us.begin2());it2 != us.end2(); ++it2){
        for(const_iterator1 it1(it2.begin());it1!=it2.end(); ++it1){
	  result.set(it1.index1(),it1.index2(),*it1);
	}
      }
    }
  };

}//of pu
#endif
